using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using System.IO;
using System.ComponentModel;

namespace AAL77FDRPartialDecoder
{
    public class SubFrameInfo
    {
        public static int[] syncWords = { 583, 1464, 2631, 3512 };

        public int subFrameNumber = 0;
        public int superFrameNumber = -1;
        public int[] subFrame = null;   // 12 bit words of subFrame data
        public int bitPosition = 0;
        public bool incomplete = false;

        //Extract subFrameNumber and superFrameNumber from subFrame
        public void GenerateInfo(bool aIncomplete)
        {
            incomplete = aIncomplete;
            int syncWordNumber = 0;
            for (; syncWordNumber < syncWords.Length; syncWordNumber++)
                if (subFrame[0] == syncWords[syncWordNumber])
                    break;
            subFrameNumber = syncWordNumber + 1;
            switch (subFrameNumber)
            {
                case 1:
                    if (superFrameNumber >= 0)
                        superFrameNumber = (superFrameNumber % 16) + 1;
                    break;
                case 3:
                    //superFrameNumber = (subFrame[63] >> 8) + 1;
                    break;
            }
        }
    }

    abstract class ParameterRow
    {
        protected int wordNum;
        protected int subFrameNumber;
        protected int superFrameNumber;
        protected int MSB;
        protected int LSB;
        protected double offset;
        protected double scalar;
        protected int decimalPlaces;

        public bool HasValue(SubFrameInfo subFrameInfo)
        {
            return (((subFrameInfo.subFrameNumber == subFrameNumber) || (subFrameNumber == 0))
                    && ((subFrameInfo.superFrameNumber == superFrameNumber) || (superFrameNumber == 0)));
        }

        protected string FormatValueAsString(int intValue)
        {
            double value = (intValue * scalar) + offset;
            string formatString = string.Format("F{0:0}", decimalPlaces);
            return value.ToString(formatString);
        }

        protected int GetUnsignedValue(int aWordNum, int aMSB, int aLSB, int[] subFrame)
        {
            int result = subFrame[aWordNum - 1];
            result &= ((1 << aMSB) - 1);
            result >>= (aLSB - 1);
            return result;
        }

        protected int GetSignedValue(int aWordNum, int aMSB, int aLSB, int[] subFrame)
        {
            int result = GetUnsignedValue(aWordNum, aMSB, aLSB, subFrame);
            int bitMask = 1 << (aMSB - aLSB);
            if ((result & bitMask) != 0)
                result -= (bitMask << 1);
            return result;
        }

        protected int GetBCDValue(int aWordNum, int aMSB, int aLSB, int[] subFrame)
        {
            int BCDValue = GetUnsignedValue(aWordNum, aMSB, aLSB, subFrame);
            int result = 0;
            for (int i = (aMSB - aLSB) & -4; i >= 0; i -= 4)
                result = (result * 10) + ((BCDValue >> i) & 0xF);
            return result;
        }

        abstract public string GetValueAsString(SubFrameInfo subFrameInfo);
    }

    class UnSignedParameterRow : ParameterRow
    {
        public UnSignedParameterRow(int aWordNum, int aSubFrameNumber, int aSuperFrameNumber, int aMSB, int aLSB, double aOffset, double aScalar, int aDecimalPlaces)
        {
            wordNum = aWordNum;
            subFrameNumber = aSubFrameNumber;
            superFrameNumber = aSuperFrameNumber;
            MSB = aMSB;
            LSB = aLSB;
            offset = aOffset;
            scalar = aScalar;
            decimalPlaces = aDecimalPlaces;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString(GetUnsignedValue(wordNum, MSB, LSB, subFrameInfo.subFrame));
        }
    }

    class SignedParameterRow : ParameterRow
    {
        public SignedParameterRow(int aWordNum, int aSubFrameNumber, int aSuperFrameNumber, int aMSB, int aLSB, double aOffset, double aScalar, int aDecimalPlaces)
        {
            wordNum = aWordNum;
            subFrameNumber = aSubFrameNumber;
            superFrameNumber = aSuperFrameNumber;
            MSB = aMSB;
            LSB = aLSB;
            offset = aOffset;
            scalar = aScalar;
            decimalPlaces = aDecimalPlaces;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString(GetSignedValue(wordNum, MSB, LSB, subFrameInfo.subFrame));
        }
    }

    class BCDParameterRow : ParameterRow
    {
        public BCDParameterRow(int aWordNum, int aSubFrameNumber, int aSuperFrameNumber, int aMSB, int aLSB, double aOffset, double aScalar, int aDecimalPlaces)
        {
            wordNum = aWordNum;
            subFrameNumber = aSubFrameNumber;
            superFrameNumber = aSuperFrameNumber;
            MSB = aMSB;
            LSB = aLSB;
            offset = aOffset;
            scalar = aScalar;
            decimalPlaces = aDecimalPlaces;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString(GetBCDValue(wordNum, MSB, LSB, subFrameInfo.subFrame));
        }
    }

    class ZeroOneParameterRow : ParameterRow
    {
        protected string zeroValue;
        protected string oneValue;

        public ZeroOneParameterRow(int aWordNum, int aSubFrameNumber, int aSuperFrameNumber, int aMSBLSB, string aZeroValue, string aOneValue)
        {
            wordNum = aWordNum;
            subFrameNumber = aSubFrameNumber;
            superFrameNumber = aSuperFrameNumber;
            MSB = aMSBLSB;
            LSB = aMSBLSB;
            zeroValue = aZeroValue;
            oneValue = aOneValue;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            if (GetUnsignedValue(wordNum, MSB, LSB, subFrameInfo.subFrame) == 0)
                return zeroValue;
            else
                return oneValue;
        }
    }

    class AltitudeParameterRow : ParameterRow
    {
        public AltitudeParameterRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 1.0;
            decimalPlaces = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString((GetSignedValue(29, 12, 5, subFrameInfo.subFrame) * 512) + GetSignedValue(30, 12, 3, subFrameInfo.subFrame));
        }
    }

    class LatitudeParameterRow : ParameterRow
    {
        public LatitudeParameterRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 180.0 / 1048576;
            decimalPlaces = 5;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString((GetSignedValue(60, 12, 3, subFrameInfo.subFrame) * 2048) + GetUnsignedValue(59, 11, 1, subFrameInfo.subFrame));
        }
    }

    class LongitudeParameterRow : ParameterRow
    {
        public LongitudeParameterRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 180.0 / 1048576;
            decimalPlaces = 5;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString((GetSignedValue(62, 12, 3, subFrameInfo.subFrame) * 2048) + GetUnsignedValue(61, 11, 1, subFrameInfo.subFrame));
        }
    }

    class RadioHeightRow : ParameterRow
    {
        public RadioHeightRow(int aSubFrameNumber)
        {
            subFrameNumber = aSubFrameNumber;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 1.0;
            decimalPlaces = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString((GetUnsignedValue(31, 12, 12, subFrameInfo.subFrame) * -8192) + (GetUnsignedValue(32, 2, 1, subFrameInfo.subFrame) * 2048) + GetUnsignedValue(31, 11, 1, subFrameInfo.subFrame));
        }
    }

    class IncompleteSubFrameRow : ParameterRow
    {
        public IncompleteSubFrameRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return subFrameInfo.incomplete.ToString();
        }
    }

    class BlockPositionRow : ParameterRow
    {
        public BlockPositionRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 1.0;
            decimalPlaces = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString(subFrameInfo.bitPosition >> 20);
        }
    }

    class BitOffsetRow : ParameterRow
    {
        public BitOffsetRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 1.0;
            decimalPlaces = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString(subFrameInfo.bitPosition & 0x7FFFF);
        }
    }

    class FlightStreamNumberRow : ParameterRow
    {
        public FlightStreamNumberRow()
        {
            subFrameNumber = 0;
            superFrameNumber = 0;
            offset = 0.0;
            scalar = 1.0;
            decimalPlaces = 0;
        }

        public override string GetValueAsString(SubFrameInfo subFrameInfo)
        {
            return FormatValueAsString((subFrameInfo.bitPosition >> 19) & 1);
        }
    }

    class ParameterColumn
    {
        private List<ParameterRow> parameterRows;
        private string columnName;

        public ParameterColumn(string aColumnName, params ParameterRow[] aParameterRows)
        {
            columnName = aColumnName;
            parameterRows = new List<ParameterRow>(aParameterRows);
        }

        public string ColumnName
        {
            get
            {
                return columnName;
            }
        }

        //Return the string representation of occurrence number lineNum of the paramater in the subFrame
        public string GetValueAsString(int lineNum, SubFrameInfo subFrameInfo)
        {
            foreach (ParameterRow parameterRow in parameterRows)
                if (parameterRow.HasValue(subFrameInfo))
                    if (lineNum-- <= 0)
                        return parameterRow.GetValueAsString(subFrameInfo);
            return "";
        }
    }

    class Parameter : IComparable<Parameter>, IEquatable<Parameter>
    {
        private List<ParameterColumn> columns;
        private string name;
        private string units;

        public Parameter(string aName, string aUnits, params ParameterColumn[] aColumns)
        {
            name = aName;
            units = aUnits;
            columns = new List<ParameterColumn>(aColumns);
        }

        public int CompareTo(Parameter parameter)
        {
            return ToString().CompareTo(parameter.ToString());
        }

        public bool Equals(Parameter parameter)
        {
            return (ToString() == parameter.ToString());
        }

        public override string ToString()
        {
            string s = name;
            s = s.Replace("^ ", "");
            s = s.Replace("^", "");
            s = s.TrimEnd(null);
            return s;
        }

        public List<ParameterColumn> Columns
        {
            get
            {
                return columns;
            }
        }

        public List<string> ParameterNames
        {
            get
            {
                List<string> parameterNames = new List<string>();
                //Build each parameterName by replacing the ^ characters in the name parameter with detail from the columns
                foreach (ParameterColumn parameterColumn in columns)
                {
                    string parameterName = name;
                    string columnName = parameterColumn.ColumnName;
                    int startPos = 0;
                    while (true)
                    {
                        int i = parameterName.IndexOf('^');
                        if (i < 0)
                            break;
                        int endPos = columnName.IndexOf(',', startPos);
                        if (endPos < 0)
                            endPos = columnName.Length;
                        parameterName = parameterName.Remove(i, 1);
                        parameterName = parameterName.Insert(i, columnName.Substring(startPos, endPos - startPos));
                        startPos = endPos + 1;
                    }
                    if (units != "")
                        parameterName += " " + units;
                    parameterNames.Add(parameterName);
                }
                return parameterNames;
            }
        }
    }

    class ParametersList : List<Parameter>
    {
        public ParametersList()
        {
            Add(new Parameter("ALTITUDE (1013.25mB)", "(FEET)",
                new ParameterColumn("", new AltitudeParameterRow())));
            Add(new Parameter("COMPUTED AIRSPEED", "(KNOTS)",
                new ParameterColumn("", new UnSignedParameterRow(96, 0, 0, 12, 3, 0.0, 1.0 / 2, 1))));
            Add(new Parameter("DME DISTANCE ^", "(NM)",
                new ParameterColumn("- LEFT", new UnSignedParameterRow(222, 3, 0, 12, 3, 0.0, 0.25, 1)),
                new ParameterColumn("- RIGHT", new UnSignedParameterRow(222, 1, 0, 12, 3, 0.0, 0.25, 1))));
            Add(new Parameter("FRAME COUNTER", "",
                new ParameterColumn("", new UnSignedParameterRow(256, 1, 0, 12, 1, 0.0, 1.0, 0))));
            Add(new Parameter("GMT ^", "",
                new ParameterColumn("HOURS", new UnSignedParameterRow(255, 1, 0, 12, 8, 0.0, 1.0, 0)),
                new ParameterColumn("MINUTES", new UnSignedParameterRow(255, 1, 0, 7, 2, 0.0, 1.0, 0)),
                new ParameterColumn("SECONDS", new UnSignedParameterRow(254, 1, 0, 12, 7, 0.0, 1.0, 0))));
            Add(new Parameter("GPS", "",
                new ParameterColumn("", new ZeroOneParameterRow(254, 2, 0, 5, "NOT WORKING", "WORKING"))));
            Add(new Parameter("GROUND PROX WARNING", "",
                new ParameterColumn("", new ZeroOneParameterRow(86, 0, 0, 1, "NOT TRUE", "TRUE"))));
            Add(new Parameter("LATERAL ACCELERATION", "(G's)",
                new ParameterColumn("",
                    new UnSignedParameterRow(35, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(99, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(163, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(227, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3))));
            Add(new Parameter("LONGITUDINAL ACCEL", "(G's)",
                new ParameterColumn("",
                    new UnSignedParameterRow(33, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(97, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(161, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3),
                    new UnSignedParameterRow(225, 0, 0, 12, 3, -532 * 0.00203659, 0.00203659, 3))));
            Add(new Parameter("PRES POSN ^", "(DEG)",
                new ParameterColumn("LAT", new LatitudeParameterRow()),
                new ParameterColumn("LONG", new LongitudeParameterRow())));
            Add(new Parameter("RADIO HEIGHT ^", "(FEET)",
                new ParameterColumn("CAPT", new RadioHeightRow(1)),
                new ParameterColumn("LRRAL", new RadioHeightRow(2)),
                new ParameterColumn("LRRAR", new RadioHeightRow(3)),
                new ParameterColumn("LRRAC", new RadioHeightRow(4))
                //,
                //new ParameterColumn("F/O", new SignedParameterRow(204, 0, 0, 12, 3, 0.0, 4.0, 0))
                ));
            Add(new Parameter("SYNC WORD", "",
                new ParameterColumn("", new UnSignedParameterRow(1, 0, 0, 12, 11, 0.0, 1.0, 0))));
            Add(new Parameter("VERTICAL ACCELERATION", "(G's)",
                new ParameterColumn("",
                    new UnSignedParameterRow(2, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(34, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(66, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(98, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(130, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(162, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(194, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3),
                    new UnSignedParameterRow(226, 0, 0, 12, 1, -1474 * 0.00228936, 0.00228936, 3))));
            Add(new Parameter("Block Position", "",
                new ParameterColumn("", new BlockPositionRow())));
            Add(new Parameter("Bit Offset", "",
                new ParameterColumn("", new BitOffsetRow())));
            Add(new Parameter("Flight Stream Number", "",
                new ParameterColumn("", new FlightStreamNumberRow())));
            Add(new Parameter("Incomplete Subframe", "",
                new ParameterColumn("", new IncompleteSubFrameRow())));
        }

        public int IndexOf(string s)
        {
            Parameter parameter = new Parameter(s, "");
            return IndexOf(parameter);
        }
    }

    //This class allows the input file to be read in units of 128 byte pages.
    public class PageFileStream
    {
        private const int blockLength = 0x20000;

        private FileStream inputFile = null;
        private int inputFileOffset = 0;
        private int inputFileStart = 0;
        private int inputFileLength = 0;
        private int inputFilePos = 0;

        public PageFileStream(string fileName)
        {
            inputFile = new FileStream(fileName, FileMode.Open, FileAccess.Read);
            int fl = (int)inputFile.Length;
            bool finished = false;
            //Search through each pair of 64KB blocks for the earliest recorded data 
            //and find out how long the data is not including the text header at the 
            //beginning of the file.
            for (int fp = fl - blockLength; (fp >= 0) && (!finished); fp -= blockLength)
            {
                inputFile.Seek(fp, SeekOrigin.Begin);
                byte[] headerBytes = new byte[2];
                inputFile.Read(headerBytes, 0, headerBytes.Length);
                int headerWord = (headerBytes[1] << 8) | headerBytes[0];
                switch (headerWord)
                {
                    case 0xFE6B:
                        //This is a pair of 64KB blocks containing valid data.
                        inputFileOffset = fp;
                        break;
                    case 0xFFFF:
                        //This is the pair of blank blocks which is left erased and ready to 
                        //be used when the current pair of blocks is full so the earliest 
                        //recorded data is the next pair of 64KB blocks.
                        inputFileOffset = fp;
                        inputFileStart = fp + blockLength;
                        break;
                    default:
                        //This is invalid data so we must be looking through the text header 
                        //at the beginning of the file.
                        finished = true;
                        break;
                }
            }
            if (inputFileStart >= fl)
                inputFileStart = inputFileOffset;
            //Calculate the start point and length of the data in units of 128 byte pages.
            inputFileStart = (inputFileStart - inputFileOffset) >> 7;
            inputFileLength = (fl - inputFileOffset) >> 7;
            Seek(0);
        }

        public int Length
        {
            get
            {
                return inputFileLength - (blockLength >> 7);
            }
        }

        public void Seek(int fp)
        {
            //Seek in to the file in units of 128 byte pages. Offset 0 is the earliest 
            //recorded data. The recorded data wraps around to the beginning of the file as
            //in a standard ring buffer.
            int pagePos = fp + inputFileStart;
            if (pagePos >= inputFileLength)
                pagePos -= inputFileLength;
            int bytePos = (pagePos << 7) + inputFileOffset;
            inputFile.Seek(bytePos, SeekOrigin.Begin);
            inputFilePos = fp;
        }

        public byte[] ReadPage()
        {
            //Read a single 128 byte page.
            byte[] page = new byte[128];
            inputFile.Read(page, 0, page.Length);
            if ((++inputFilePos + inputFileStart) == inputFileLength)
                Seek(inputFilePos);
            return page;
        }

        public void Close()
        {
            inputFile.Close();
        }
    }

    //This class is for sorting the output in ascending order of Block Position, Bit Offset,
    //then Flight Stream Number. This is necessary since each pair of 64KB blocks (which have
    //the same Block Position) is filled simultaneously. The first block in each pair is from
    //Flight Stream Number 0 ane the second block in each pair is from Flight Stream Number 1.
    //By sorting by Bit Offset we end up with the data being output in approximately but not
    //exactly in the same order in which it was recorded.
    class OutputKey : IComparable<OutputKey>
    {
        private int blockPosition;
        private int bitOffset;
        private int flightStreamNumber;
        private int lineNumber;

        public OutputKey(int bitPosition, int aLineNumber)
        {
            blockPosition = bitPosition >> 20;
            bitOffset = bitPosition & 0x7FFFF;
            flightStreamNumber = (bitPosition >> 19) & 1;
            lineNumber = aLineNumber;
        }

        public int CompareTo(OutputKey outputKey)
        {
            if (blockPosition == outputKey.blockPosition)
                if (bitOffset == outputKey.bitOffset)
                    if (flightStreamNumber == outputKey.flightStreamNumber)
                        return lineNumber - outputKey.lineNumber;
                    else
                        return flightStreamNumber - outputKey.flightStreamNumber;
                else
                    return bitOffset - outputKey.bitOffset;
            else
                return blockPosition - outputKey.blockPosition;
        }
    }

    class GenerateOutput
    {
        const int subFrameLength = 256;

        private Main main = null;
        private SortedDictionary<OutputKey, string> output = new SortedDictionary<OutputKey, string>();
        private int flightStreamNumber = 0;
        private SubFrameInfo queuedSubFrameInfo = new SubFrameInfo();
        private int consecutiveSyncWords = 0;
        private int subFrameStart = 0;
        private int rowNum = 0;

        private void RecordHeader()
        {
            string line = "Bit Position";
            foreach (Parameter parameter in main.includedParametersListBox.Items)
                foreach (string parameterName in parameter.ParameterNames)
                    line += "," + parameterName;
            main.outputFile.WriteLine(line);
            rowNum = 1;
        }

        private void RecordSubFrame(bool incomplete)
        {
            if (queuedSubFrameInfo.subFrame != null)
            {
                queuedSubFrameInfo.GenerateInfo(incomplete);
                for (int lineNum = 0; ; lineNum++)
                {
                    bool emptyLine = true;
                    string line = queuedSubFrameInfo.bitPosition.ToString();
                    foreach (Parameter parameter in main.includedParametersListBox.Items)
                        foreach (ParameterColumn parameterColumn in parameter.Columns)
                        {
                            string value = parameterColumn.GetValueAsString(lineNum, queuedSubFrameInfo);
                            if (value != "")
                                emptyLine = false;
                            line += "," + value;
                        }
                    if ((!emptyLine) 
                        && ((!incomplete) || main.incompleteSubFramesCheckBox.Checked)
                        && ((lineNum <= 0) || (!main.maximumOneLinePerSubFrameCheckBox.Checked)))
                            //Store the line in sorted order to be written out when the reading process is complete.
                            output.Add(new OutputKey(queuedSubFrameInfo.bitPosition, lineNum), line);
                    if (emptyLine)
                        break;
                    }
                queuedSubFrameInfo.subFrame = null;
            }
            //If synchronisation is lost, use the superFrameNumber again only when a new one is read
            if (incomplete)
                queuedSubFrameInfo.superFrameNumber = -1;
        }

        //Information about the format in which the FDR records the data can be found in the post 
        //at http://forums.randi.org/showpost.php?p=2022420&postcount=85 and in figure 4.2-2 of 
        //the document http://www.asc.gov.tw/author_files/ASC-TRT-99-007.pdf

        //The words in the figure are 16 bit words stored Least Significant Byte first and the 
        //Hamming Code & Page Parity occupies the last 11 bits of each page. Within each 16 bit word, 
        //the most significant bit is the first bit in the bit stream. Some pages are blank 
        //consisting entirely of 0xFFFF words and need to be skipped. 

        //Up to 4 subframes of uncompressed data is generated in the file each time the FDR is 
        //started up which I believe happens when the engines are started. After that, data is 
        //stored in a compressed form which this program can not decode. The last subframe of 
        //uncompressed data recorded before the compressed data may not be complete. 
        
        //The uncompressed data is built up by taking the first bit in the bit stream as the least 
        //significant bit of a 12 bit word and then building up subframes of 256 * 12 bit words each.
        //Uncompressed subframes can be recognised by the sync word which is the first word of each 
        //subframe. The valid sync word values are listed in the SubFrameInfo.syncWords array. The 
        //program takes two consecutive sync words in the correct positions to be an indication of 
        //uncompressed data. This method is not fool proof however and for some parts of the file it 
        //tries to decode compressed data as if it was uncompressed.
        public bool Start(Main aMain, BackgroundWorker generateOutputBackgroundWorker)
        {
            main = aMain;
            int[] subFrame = new int[subFrameLength];
            int lastPagePos = -1;
            //Read one flight stream then the other.
            for (flightStreamNumber = 0; flightStreamNumber < 2; flightStreamNumber++)
            {
                bool haveSync = false;
                int lookAheadSync = -1;
                int lookAheadBitCount = 0;
                int saveW12 = 0;
                int saveBitPos = 0;
                bool foundFrameMarker = false;
                consecutiveSyncWords = 0;
                int bitPos = (flightStreamNumber << 19) + 0x800;
                int subFrameBitPos = 0;
                subFrameStart = 0;
                byte[] page = null;
                int w12 = 0;
                int w16 = 0;
                int lastWordPos = -1;
                int highestPercentageReached = 0;
                while (true)
                {
                    if (generateOutputBackgroundWorker.CancellationPending)
                    {
                        main.outputFile.WriteLine("Generation of output cancelled");
                        return (false);
                    }
                    if (((bitPos & 0x3FF) >= 1013) || (page == null))
                    {
                        if (page == null)
                            page = new byte[128];
                        else
                            //Move to the next page in this block.
                            bitPos = (bitPos & -0x400) + 0x400;
                        bool emptyPage = true;
                        while (true)
                        {
                            if ((bitPos & 0x7FFFF) == 0)
                                //Move to the next block in this flight stream.
                                bitPos += 0x80800;
                            int pagePos = (bitPos >> 10);
                            if (pagePos >= main.inputFile.Length)
                                break;
                            if (pagePos != lastPagePos + 1)
                                main.inputFile.Seek(pagePos);
                            page = main.inputFile.ReadPage();
                            lastPagePos = pagePos;
                            for (int bp = 0; bp < page.Length; bp++)
                                if (page[bp] != 0xFF)
                                {
                                    emptyPage = false;
                                    break;
                                }
                            if (emptyPage)
                                //This page is empty. Ignore it, and read the next one.
                                bitPos += 0x400;
                            else
                                break;
                        }
                        if (emptyPage)
                            break;
                    }
                    int wordPos = (bitPos >> 4) & 0x3F;
                    if (wordPos != lastWordPos)
                    {
                        //Construct a 16 bit word from the page.
                        w16 = (page[(wordPos << 1) | 1] << 8) | page[(wordPos << 1)];
                        lastWordPos = wordPos;
                        if ((consecutiveSyncWords > 0) && (w16 == 0xFEFE) && (wordPos != 63))
                            //If a frame marker is found, set the rest of the subframe to zeroes.
                            foundFrameMarker = true;
                        if (foundFrameMarker)
                            w16 = 0;
                    }
                    //Construct a 12 bit word from the 16 bit words.
                    w12 = (w12 >> 1) | (((w16 << (bitPos & 15)) & 0x8000) >> 4);
                    subFrameBitPos++;
                    if (lookAheadSync < 0)
                    {
                        //Not in sync word look ahead mode.
                        if ((!haveSync) && (subFrameBitPos >= 12))
                        {
                            //We could now have a sync word.
                            if (subFrameBitPos == 13)
                            {
                                //There was no sync word where we expected it, so the previous subFrame is incomplete and synchronisation has been lost.
                                RecordSubFrame(true);
                                consecutiveSyncWords = 0;
                                foundFrameMarker = false;
                            }
                            for (int i = 0; i < SubFrameInfo.syncWords.Length; i++)
                                if (w12 == SubFrameInfo.syncWords[i])
                                {
                                    //We found a sync word.
                                    consecutiveSyncWords++;
                                    if (subFrameBitPos > 12)
                                    {
                                        //Look ahead to see if the next sync word appears where we expect it.
                                        saveW12 = w12;
                                        saveBitPos = bitPos;
                                        lookAheadSync = SubFrameInfo.syncWords[(i + 1) % SubFrameInfo.syncWords.Length];
                                        lookAheadBitCount = 3072;
                                    }
                                    else
                                    {
                                        //We still have synchronisation and can build up the subframe.
                                        RecordSubFrame(false);
                                        subFrameBitPos = 12;
                                        haveSync = true;
                                    }
                                    subFrameStart = bitPos;
                                    break;
                                }
                        }
                    }
                    else
                        //In sync word look ahead mode.
                        if (--lookAheadBitCount <= 0)
                        {
                            if (haveSync = (w12 == lookAheadSync))
                                //Look ahead successful. We now have synchronsiation and can build up the subframe.
                                subFrameBitPos = 12;
                            else
                                //Look ahead failed. Try again.
                                consecutiveSyncWords = 0;
                            foundFrameMarker = false;
                            page = null;
                            lastWordPos = -1;
                            w12 = saveW12;
                            bitPos = saveBitPos;
                            lookAheadSync = -1;
                        }
                    if ((haveSync) && ((subFrameBitPos % 12) == 0))
                    {
                        //We now have the complete 12 bits of a word, so save it.
                        int subFrameWordPos = subFrameBitPos / 12;
                        subFrame[subFrameWordPos - 1] = w12;
                        if (subFrameWordPos >= subFrame.Length)
                        {
                            //We now have a complete subframe.                            
                            queuedSubFrameInfo.subFrame = subFrame;
                            queuedSubFrameInfo.bitPosition = subFrameStart;
                            subFrame = new int[subFrame.Length];
                            haveSync = false;
                            subFrameBitPos = 0;
                        }
                    }
                    if ((bitPos & 0xFF) == 0)
                    {
                        int percentComplete = 50 * flightStreamNumber + (int)(lastPagePos * 50.0 / main.inputFile.Length);
                        if (percentComplete > highestPercentageReached)
                        {
                            highestPercentageReached = percentComplete;
                            generateOutputBackgroundWorker.ReportProgress(percentComplete);
                        }
                    }
                    bitPos++;
                }
            }
            //Write out the sorted file.
            RecordHeader();
            foreach (string line in output.Values)
            {
                main.outputFile.WriteLine(line);
                rowNum++;
            }
            return (true);
        }
    }
}
